@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.6.2.9 2009/08/22 04:19:49 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.6.2.17 2009/11/29 08:29:14 t-ishii Exp $
  * 
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -68,8 +68,14 @@ int execute_select = 0; /* non 0 if select query is in transaction */
 
 /* non 0 if "BEGIN" query with extended query protocol received */
 int receive_extended_begin = 0;
-/* non 0 if allow to close internal transaction */
-int allow_close_transaction = 1;
+
+/*
+ * Non 0 if allow to close internal transaction.  This variable was
+ * introduced on 2008/4/3 not to close an internal transaction when
+ * Sync message is received after receiving Parse message. This hack
+ * is for PHP-PDO.
+ */
+static int allow_close_transaction = 1;
 
 PreparedStatementList prepared_list; /* prepared statement name list */
 
@@ -90,6 +96,7 @@ char *parsed_query = NULL;
 
 static int check_errors(POOL_CONNECTION_POOL *backend, int backend_id);
 static void generate_error_message(char *prefix, int specific_error, char *query);
+static int is_temp_table(POOL_CONNECTION_POOL *backend, Node *node);
 
 POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 										POOL_CONNECTION_POOL *backend)
@@ -477,14 +484,14 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 		if (REPLICATION)
 		{
 			/* start a transaction if needed */
-			if (start_internal_transaction(backend, (Node *)node) != POOL_CONTINUE)
+			if (start_internal_transaction(frontend, backend, (Node *)node) != POOL_CONTINUE)
 				return POOL_END;
 
 			/* check if need lock */
 			if (need_insert_lock(backend, string, node))
 			{
 				/* if so, issue lock command */
-				status = insert_lock(backend, string, (InsertStmt *)node);
+				status = insert_lock(frontend, backend, string, (InsertStmt *)node);
 				if (status != POOL_CONTINUE)
 				{
 					free_parser();
@@ -492,7 +499,7 @@ POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend,
 				}
 			}
 		}
-		else if (REPLICATION && query == NULL && start_internal_transaction(backend, node))
+		else if (REPLICATION && query == NULL && start_internal_transaction(frontend, backend, node))
 		{
 			free_parser();
 			return POOL_ERROR;
@@ -669,22 +676,30 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 		p_stmt = (PrepareStmt *)portal->stmt;
 
 		string1 = portal->sql_string;
+		pool_debug("Execute: query: %s", string1);
 		node = (Node *)p_stmt->query;
 		strncpy(query_string_buffer, string1, sizeof(query_string_buffer));
 
-		if ((IsA(node, PrepareStmt) || IsA(node, DeallocateStmt) ||
-			 IsA(node, VariableSetStmt)) &&
-			MASTER_SLAVE && TSTATE(backend) != 'E')
+ 		if ((IsA(node, PrepareStmt) || IsA(node, DeallocateStmt) ||
+ 			 IsA(node, VariableSetStmt)) &&
+  			MASTER_SLAVE && TSTATE(backend) != 'E')
 		{
+			/*
+			 * PREPARE, DEALLOCATE, SET, DISCARD
+			 * should be executed on all nodes.  So we set
+			 * force_replication.
+			 */
 			force_replication = 1;
 		}
 		/*
-		 * JDBC driver sends "BEGIN" query internally if setAutoCommit(false).
-		 * But it does not send Sync message after "BEGIN" query.
-		 * In extended query protocol, PostgreSQL returns
-		 * ReadyForQuery when a client sends Sync message.
-		 * We can't know a transaction state...
-		 * So pgpool send Sync message internally.
+		 * JDBC driver sends "BEGIN" query internally if
+		 * setAutoCommit(false).  But it does not send Sync message
+		 * after "BEGIN" query.  In extended query protocol,
+		 * PostgreSQL returns ReadyForQuery when a client sends Sync
+		 * message.  Problem is, pgpool can't know the transaction
+		 * state without receiving ReadyForQuery. So we remember that
+		 * we need to send Sync message internally afterward, whenever
+		 * we receive BEGIN in extended protocol.
 		 */
 		else if (IsA(node, TransactionStmt) && MASTER_SLAVE)
 		{
@@ -692,7 +707,8 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend,
 
 			if (stmt->kind == TRANS_STMT_BEGIN ||
 				stmt->kind == TRANS_STMT_START)
-				receive_extended_begin = true;
+				/* Remember we need to send sync later in extended protocol */
+				receive_extended_begin = 1;
 		}
 
 		if (load_balance_enabled(backend, node, string1))
@@ -907,10 +923,32 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
 	}
 	else
 	{
+		/* Save last query string for logging purpose */
+		snprintf(query_string_buffer, sizeof(query_string_buffer), "Parse: %s", stmt);
+
 		node = (Node *) lfirst(list_head(parse_tree_list));
 
 		insert_stmt_with_lock = need_insert_lock(backend, stmt, node);
 
+		/* Special treatment for master/slave + temp tables */
+		if (MASTER_SLAVE)
+		{
+			/* Is there "NO LOAD BALANCE" comment? */
+			if (!strncasecmp(stmt, NO_LOAD_BALANCE, NO_LOAD_BALANCE_COMMENT_SZ) ||
+				/* or the table used in a query is a temporary one ? */
+				is_temp_table(backend, node))
+			{
+				/*
+				 * From now on, let only master handle queries.  This is
+				 * typically usefull for using temp tables in master/slave
+				 * mode
+				 */
+				master_slave_was_enabled = 1;
+				MASTER_SLAVE = 0;
+				master_slave_dml = 1;
+			}
+		}
+
 		portal = create_portal();
 		if (portal == NULL)
 		{
@@ -971,12 +1009,12 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
 			}
 
 			if (is_strict_query(node))
-				start_internal_transaction(backend, node);
+				start_internal_transaction(frontend, backend, node);
 
 			if (insert_stmt_with_lock)
 			{
 				/* start a transaction if needed and lock the table */
-				status = insert_lock(backend, stmt, (InsertStmt *)node);
+				status = insert_lock(frontend, backend, stmt, (InsertStmt *)node);
 				if (status != POOL_CONTINUE)
 				{
 					return status;
@@ -993,12 +1031,22 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
 
 	if (REPLICATION || PARALLEL_MODE || MASTER_SLAVE)
 	{
-		/* We must synchronize because Parse message acquires table
+		/*
+		 * We must synchronize because Parse message acquires table
 		 * locks.
 		 */
-		pool_debug("waiting for master completing the query");
-		if (synchronize(MASTER(backend)))
+		pool_debug("Parse: waiting for master completing the query");
+		if (wait_for_query_response(frontend, MASTER(backend), string, MAJOR(backend)) != POOL_CONTINUE)
+		{
+			/* Cancel current transaction */
+			CancelPacket cancel_packet;
+
+			cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+			cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+			cancel_packet.key= MASTER_CONNECTION(backend)->key;
+			cancel_request(&cancel_packet);
 			return POOL_END;
+		}
 
 		/*
 		 * We must check deadlock error because a aborted transaction
@@ -1036,9 +1084,18 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend,
 			if (!VALID_BACKEND(i) || IS_MASTER_NODE_ID(i))
 				continue;
 
-			pool_debug("waiting for %dth backend completing the query", i);
-			if (synchronize(CONNECTION(backend, i)))
+			pool_debug("Parse: waiting for %dth backend completing the query", i);
+			if (wait_for_query_response(frontend, CONNECTION(backend, i), string, MAJOR(backend)) != POOL_CONTINUE)
+			{
+				/* Cancel current transaction */
+				CancelPacket cancel_packet;
+
+				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+				cancel_packet.key= MASTER_CONNECTION(backend)->key;
+				cancel_request(&cancel_packet);
 				return POOL_END;
+			}
 		}
 	}
 
@@ -1163,7 +1220,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			pool_debug("ReadyForQuery: transaction state: %c", state);
 		}
 
-		if (end_internal_transaction(backend) != POOL_CONTINUE)
+		if (end_internal_transaction(frontend, backend) != POOL_CONTINUE)
 			return POOL_ERROR;
 	}
 
@@ -1226,16 +1283,14 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			pool_write(frontend, &len, sizeof(len));
 			pool_write(frontend, &state, 1);
 		}
-
-		if (pool_flush(frontend))
-			return POOL_END;
+		pool_flush(frontend);
 	}
 
 	in_progress = 0;
 
 	/* end load balance mode */
 	if (in_load_balance)
-		end_load_balance(backend);
+		end_load_balance();
 
 	if (master_slave_dml)
 	{
@@ -1445,6 +1500,10 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 
 	pool_debug("read kind from frontend %c(%02x)", fkind, fkind);
 
+	/*
+	 * If we have received BEGIN in extended protocol before, we need
+	 * to send a sync message to know the transaction stare.
+	 */
 	if (receive_extended_begin)
 	{
 		receive_extended_begin = 0;
@@ -1483,10 +1542,20 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 
 		case 'P':  /* Parse message */
 			allow_close_transaction = 0;
+
+			if (MASTER_SLAVE &&
+				(TSTATE(backend) != 'I' || receive_extended_begin))
+			{
+				pool_debug("kind: %c master_slave_dml enabled", fkind);
+				master_slave_was_enabled = 1;
+				MASTER_SLAVE = 0;
+				master_slave_dml = 1;
+			}
+
 			status = Parse(frontend, backend);
 			break;
 
-		case 'S':
+		case 'S':  /* Sync message */
 			receive_extended_begin = 0;
 			/* fall through */
 
@@ -1501,7 +1570,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 					MASTER_SLAVE = 0;
 					master_slave_dml = 1;
 				}
-
+	
 				status = SimpleForwardToBackend(fkind, frontend, backend);
 				for (i=0;i<NUM_BACKENDS;i++)
 				{
@@ -2374,3 +2443,106 @@ static void generate_error_message(char *prefix, int specific_error, char *query
 	pool_error(msg->data, query);
 	free_string(msg);
 }
+
+/*
+ * Judge the table used in a query represented by node is a temporary
+ * table or not.
+ */
+static int is_temp_table(POOL_CONNECTION_POOL *backend, Node *node)
+{
+/*
+ * Query to know if pg_class has relistemp column or not.
+ * PostgreSQL 8.4 or later has this.
+ */
+#define HASRELITEMPPQUERY "SELECT count(*) FROM pg_catalog.pg_class AS c, pg_attribute AS a WHERE c.relname = 'pg_class' AND a.attrelid = c.oid AND a.attname = 'relistemp'"
+
+/*
+ * Query to know if the target table is a temporary one.
+ * This query is valid through PostgreSQL 7.3 to 8.3.
+ */
+#define ISTEMPQUERY83 "SELECT count(*) FROM pg_class AS c, pg_namespace AS n WHERE c.relname = '%s' AND c.relnamespace = n.oid AND n.nspname ~ '^pg_temp_'"
+
+/*
+ * Query to know if the target table is a temporary one.
+ * This query is valid PostgreSQL 8.4 or later.
+ */
+#define ISTEMPQUERY84 "SELECT count(*) FROM pg_catalog.pg_class AS c WHERE c.relname = '%s' AND c.relistemp"
+
+	char *str;
+	int hasrelistemp;
+	int result;
+	static POOL_RELCACHE *hasrelistemp_cache;
+	static POOL_RELCACHE *relcache;
+	char *query;
+
+	/*
+	 * For version 2 protocol, we cannot support the checking since
+	 * the underlying infrastructure (do_query) does not support the
+	 * protocol. So we just return false.
+	 */
+	if (MAJOR(backend) == PROTO_MAJOR_V2)
+		return 0;
+
+	/* For SELECT, it's hard to extract table names. So we always return 0 */
+	if (IsA(node, SelectStmt))
+	{
+		return 0;
+	}
+
+	/* Obtain table name */
+	if (IsA(node, InsertStmt))
+		str = nodeToString(((InsertStmt *)node)->relation);
+	else if (IsA(node, UpdateStmt))
+		str = nodeToString(((UpdateStmt *)node)->relation);
+	else if (IsA(node, DeleteStmt))
+		str = nodeToString(((DeleteStmt *)node)->relation);
+	else		/* Unknown statement */
+		str = NULL;
+
+	if (str == NULL)
+	{
+			return 0;
+	}
+
+	/*
+	 * Check backend version
+	 */
+	if (!hasrelistemp_cache)
+	{
+		hasrelistemp_cache = pool_create_relcache(32, HASRELITEMPPQUERY,
+										int_register_func, int_unregister_func,
+										false);
+		if (hasrelistemp_cache == NULL)
+		{
+			pool_error("is_temp_table: pool_create_relcache error");
+			return false;
+		}
+	}
+
+	hasrelistemp = (int)pool_search_relcache(hasrelistemp_cache, backend, "pg_class");
+	if (hasrelistemp)
+		query = ISTEMPQUERY84;
+	else
+		query = ISTEMPQUERY83;
+
+	/*
+	 * If relcache does not exist, create it.
+	 */
+	if (!relcache)
+	{
+		relcache = pool_create_relcache(32, query,
+										int_register_func, int_unregister_func,
+										true);
+		if (relcache == NULL)
+		{
+			pool_error("is_temp_table: pool_create_relcache error");
+			return false;
+		}
+	}
+
+	/*
+	 * Search relcache.
+	 */
+	result = (int)pool_search_relcache(relcache, backend, str);
+	return result;
+}
